unit NumEdit;

{ Copyright (C) 1997 by Mario Ferrari }

{
  TNumEdit: Edit box per input numerico (il controllo non gestisce notazione
            esponenziale).

  Propriet speciali
  ------------------
  DecimalDigits:    Numero di decimali ammessi (0-9). E' questo il campo che
                    determina l'effettivo numero di decimali ammessi durante
                    l'input da tastiera e restituiti in output dalla propriet
                    Value. L'assegnazione di questa propriet determina anche
                    la riassegnazione automatica di un DisplayFormat di default.
  DisplayFormat:    Formato di visualizzazione dei valori quando il campo non
                    ha focus. A rischio e pericolo di chi lo usa, pu anche
                    non essere coerente con il numero di decimali ammessi
                    (sconsigliato). Durante l'edit dei dati, questi sono sempre
                    mostrati in formato "piatto": eventuale segno meno + parte
                    intera (senza punti migliaia) + eventuale punto decimale +
                    eventuale parte decimale.
  MaxValue:         Valore massimo ammesso per il campo (v.RangeErrorMsg).
  MinValue:         Valore minimo ammesso per il campo (v.RangeErrorMsg).
  RangeErrorMsg:    Messaggio di errore da visualizzare quando si esegue il
                    metodo Validate se il valore inserito non  incluso tra
                    MinValue e MaxValue.
  Value:            Valore numerico del dato inserito nel controllo (extended).
                    Attenzione: il numero di decimali  sempre quello specifi-
                    cato da DecimalDigits, anche se l'utente non ha effettuato
                    input ed il dato  stato solo assegnato e poi riletto dal
                    controllo. Dato che l'eventuale troncamento avviene
                    durante l'assegnazione del dato, la propriet Value deve
                    sempre essere assegnata dopo DecimalDigits.

  Metodi speciali
  ---------------
  Validate:         Forza il controllo di range. Se il valore non  valido,
                    restituisce false e visualizza il messaggio di errore.
}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Menus, Forms, Dialogs, StdCtrls, ExtCtrls, Buttons;

type
  TNumEdit = class(TCustomMemo)
  private
    FDecDigits: byte;
    FDispFormat: string;
    FValue: extended;
    FMaxValue: extended;
    FMinValue: extended;
    FRangeErrorMsg: string;
    procedure SetDecimalDigits(b: byte);
    procedure SetDisplayFormat(s: string);
    procedure SetMaxValue(e: extended);
    procedure SetMinValue(e: extended);
    procedure SetRangeErrorMsg(s: string);
    procedure SetValue(e: extended);
    procedure FormatText;
    procedure UnFormatText;
    function  UpdateValue: string;
  protected
    procedure Change; override;
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure KeyPress(var Key: char); override;
    procedure CreateParams(var Params: TCreateParams); override;
  public
    constructor Create(AOwner: TComponent); override;
    function Validate: boolean;
  published
    property Alignment default taRightJustify;
    property AutoSelect;
    property AutoSize default True;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DecimalDigits: byte read FDecDigits write SetDecimalDigits;
    property DisplayFormat: string read FDispFormat write SetDisplayFormat;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property MaxLength;
    property MaxValue: extended read FMaxValue write SetMaxValue;
    property MinValue: extended read FMinValue write SetMinValue;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RangeErrorMsg: string read FRangeErrorMsg write SetRangeErrorMsg;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property Value: extended read FValue write SetValue;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Others', [TNumEdit]);
end;

constructor TNumEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Alignment := taRightJustify;
  AutoSelect := true;
  AutoSize := false;
  DecimalDigits := 0;
  DisplayFormat := ',0';
  Height := 21;
  MaxValue := 1e99;
  MinValue := -1e99;
  RangeErrorMsg := '';
  Value := 0.0;
  WantReturns := false;
  Width := 121;
  WordWrap := False;
  FormatText;
end;

procedure TNumEdit.SetDecimalDigits(b: byte);
var tmp: string; i: byte;
begin
  if b<0 then b:=0;
  if b>9 then b:=9;
  if FDecDigits <> b then begin
    FDecDigits := b;
    if b=0 then
      tmp := ',0'
    else begin
      tmp := ',0.';
      for i:=1 to b do tmp:=tmp+'0';
    end;
    DisplayFormat := tmp;
    FormatText;
  end;
end;

procedure TNumEdit.SetDisplayFormat(s: string);
begin
  if FDispFormat <> s then begin
    FDispFormat:= s;
    FormatText;
  end;
end;

procedure TNumEdit.SetMaxValue(e: extended);
begin
  if FMaxValue <> e then FMaxValue := e;
end;

procedure TNumEdit.SetMinValue(e: extended);
begin
  if FMinValue <> e then FMinValue := e;
end;

procedure TNumEdit.SetValue(e: extended);
begin
  if FValue <> e then begin
    FValue := e;
    FormatText;
    if Focused then UnformatText;
  end;
end;

procedure TNumEdit.SetRangeErrorMsg(s: string);
begin
  if FRangeErrorMsg <> s then FRangeErrorMsg := s;
end;

function TNumEdit.UpdateValue: string;
var
  sTmp: string; i: byte;
begin
  if (Pos('-',Text) > 0) or (Pos('(',Text) > 0) then
    sTmp := '-'
  else
    sTmp := '';
  for i := 1 to Length(Text) do
    if Text[i] in ['0'..'9',DecimalSeparator] then
      sTmp := sTmp + Text[i];
  i := Pos(DecimalSeparator,Text);
  if (i>0) and (Length(sTmp)-i > FDecDigits) then Delete(sTmp,i+FDecDigits,255);
  if sTmp='' then sTmp:='0';
  try
    FValue := StrToFloat(sTmp);
  except
    MessageBeep(mb_IconExclamation);
  end;
  Result := sTmp;
end;

procedure TNumEdit.UnFormatText;
var
  sTmp: string;
begin
  sTmp := UpdateValue;
  if FValue<>0 then Text:=sTmp else Text:='';
end;

procedure TNumEdit.FormatText;
begin
  Text := FormatFloat(FDispFormat,FValue);
end;

function TNumEdit.Validate: boolean;
begin
  UnformatText;
  FormatText;
  if (FRangeErrorMsg>'')
  and ((FValue>FMaxValue) or (FValue<FMinValue)) then begin
    MessageBeep(mb_IconExclamation);
    MessageDlg(FRangeErrorMsg,mtError,[mbOK],0);
    SetFocus;
    Result := false;
  end;
  Result := true;
end;

procedure TNumEdit.Change;
begin
  UpdateValue;
  inherited Change;
end;

procedure TNumEdit.DoEnter;
begin
  UnformatText;
  if AutoSelect then SelectAll;
  inherited DoEnter;
end;

procedure TNumEdit.DoExit;
begin
  UnformatText;
  FormatText;
  inherited DoExit;
end;

procedure TNumEdit.KeyPress(var Key: Char);
var i: byte;
begin
  case Key of
    '0'..'9': begin
                if SelLength>0 then SelText:='';
                i := Pos(DecimalSeparator,Text);
                if (i>0) and (SelStart>=i) and (Length(Text)-i >= FDecDigits) then Key:=#0;
                if (SelStart=0) and (copy(Text,1,1)='-') then Key:=#0;
              end;
    '.',',':  begin
                if SelLength>0 then SelText:='';
                if (Pos(DecimalSeparator,Text)=0)
                and (FDecDigits>0) then
                  Key:=DecimalSeparator
                else
                  Key:=#0;
              end;
    '-'     : begin
                if SelLength>0 then SelText:='';
                if (Pos('-',Text)>0) or (SelStart>0) then Key:=#0;
              end;
    chr(VK_BACK):;  {OK}
  else
    Key := #0;
  end;
  inherited KeyPress(Key);
end;

procedure TNumEdit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  case Alignment of
    taLeftJustify : Params.Style := Params.Style or ES_LEFT and Not ES_MULTILINE;
    taRightJustify: Params.Style := Params.Style or ES_RIGHT and Not ES_MULTILINE;
    taCenter      : Params.Style := Params.Style or ES_CENTER and Not ES_MULTILINE;
  end;
end;

end.
